--imports
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Parser = require("dogma.syn.Parser")
local SentType = require("dogma.syn.SentType")
local StmtType = require("dogma.syn.StmtType")

--Suite.
return suite("dogma.syn.StmtParser", function()
  local parser, stmt

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
  end):title("Create parser")

  ------------
  -- common --
  ------------
  suite("common", function()
    test("x = 12 + 34", function(params)
      parser:parse(params[1] .. " x = 12 + 34")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="std", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 12 34)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("several declarations", function(params)
      parser:parse(params[2])
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(2)
      assert(stmt.decls[1]):has({type=params[1], subtype="std", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 12 34)")
      assert(stmt.decls[2]):has({type=params[1], subtype="std", names={"y"}})
      assert(tostring(stmt.decls[2].value)):eq("(* 56 78)")
    end):iter(
      {subtitle=" # var ...", params={"var", "var x = 12+34, y = 56*78"}},
      {subtitle=" # var (...)", params={"var", "var (\n  x = 12+34, y = 56*78\n)"}},
      {subtitle=" # const ...", params={"const", "const x = 12+34, y = 56*78"}},
      {subtitle=" # const (...)", params={"const", "const (\n  x = 12+34, y = 56*78\n)"}}
    )

    test("error - comma expected", function(params)
      parser:parse(params[1] .. " x = 12 ; y")
      assert(function() parser:next() end):raises("',' expected.")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("error - comma expected", function(params)
      parser:parse(params[1] .. " (\nx = 12;\ny\n)")
      assert(function() parser:next() end):raises("',' or end of line expected.")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("error - incomplete statement", function(params)
      parser:parse(params[1])
      assert(function() parser:next() end):raises("incomplete const/var.")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )
  end)

  ---------
  -- var --
  ---------
  suite("var", function()
    test("var x", function()
      parser:parse("var x")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type="var", subtype="std", names={"x"}, value=nil})
    end)

    test("visibility var x", function(params)
      parser:parse(params[1] .. " var x")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR, visib=params[1]})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type="var", subtype="std", names={"x"}, value=nil})
    end):iter(
      {subtitle=" # export", params="export"},
      {subtitle=" # pub", params="pub"}
    )

    test("var name ::= exp", function()
      parser:parse("var x ::= 12+34")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type="const", subtype="std", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 12 34)")
    end)
  end)

  -----------
  -- const --
  -----------
  suite("const", function()
    test("error - equal expected", function()
      parser:parse("const x")
      assert(function() parser:next() end):raises("anonymous code (1, 8): '=' expected.")
    end)

    test("visibility const x = 12", function(params)
      parser:parse(params[1] .. " const x = 12")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR, visib=params[1]})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type="const", subtype="std", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("12")
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )
  end)

  ---------
  -- map --
  ---------
  suite("map", function()
    test("{name} = exp", function(params)
      parser:parse(params[1] .. " {x} = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="map", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle= "# const", params="const"}
    )

    test("{name, name} = exp", function(params)
      parser:parse(params[1] .. " {x, y} = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="map", names={"x", "y"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("var {name, name} ::= exp", function(params)
      parser:parse("var {x, y} ::= 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type="const", subtype="map", names={"x", "y"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end)
  end)

  ----------
  -- list --
  ----------
  suite("list", function()
    test("[name] = exp", function(params)
      parser:parse(params[1] .. " [x] = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="list", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("var [name] ::= exp", function()
      parser:parse("var [x] ::= 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type="const", subtype="list", names={"x"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end)

    test("[name, name] = exp", function(params)
      parser:parse(params[1] .. " [x, y] = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="list", names={"x", "y"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("[_, name] = exp", function(params)
      parser:parse(params[1] .. " [_ ,y] = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="list", names={"", "y"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("[name, _,name] = exp", function(params)
      parser:parse(params[1] .. " [x, _, z] = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="list", names={"x", "", "z"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("[name, ...name] = exp", function(params)
      parser:parse(params[1] .. " [x, ...y] = 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.VAR})
      assert(stmt.decls):len(1)
      assert(stmt.decls[1]):has({type=params[1], subtype="list", names={"x", "...y"}})
      assert(tostring(stmt.decls[1].value)):eq("(+ 1 2)")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )

    test("error - name expected", function(params)
      parser:parse(params[1] .. " [x, , y] = 1 + 2")
      assert(function() parser:next() end):raises("name expected.")
    end):iter(
      {subtitle=" # var", params="var"},
      {subtitle=" # const", params="const"}
    )
  end)
end)
